import os

PROJECT_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
ENVIRONMENT = os.environ.get("ENVIRONMENT")
DB_USERNAME = os.environ.get(f"DB_USERNAME")
DB_PASSWORD = os.environ.get(f"DB_PASSWORD")
DB_HOSTNAME = os.environ.get(f"DB_HOSTNAME").replace("env", ENVIRONMENT)


class ExampleModel():
    """Example business logic that can be wrapped into a model.
       The class _must_ contain a 'predict' method.
    """
    def business_logic(self, record: dict) -> bool:
        if record.get("sales") > 5:
            return True
        else:
            return False

    def predict(self, model_input: dict) -> bool:
        return self.business_logic()


if __name__ == "__main__":
    """
    Saves the model to MLflow in an experiment run
    The model can be registered as a new version within the MLflow UI
    """
    from umlaut.core import Umlaut

    model = Umlaut(model_name="example model")
    model.track_model(ExampleModel())



import os
from typing import Any
import mlflow
import pandas as pd
import json
from contextlib import suppress


class PyfuncWrapper(mlflow.pyfunc.PythonModel):
    """Reusable MLflow pyfunc wrapper"""

    def __init__(self, model):
        """Map variables from model to wrapper
        :model class: The class object of the model to be wrapped and saved in MLflow
        :figures dict: A dictionary of {'<filename>': Plotly graph object} to be logged
        for model performance visualization
        """
        self.model = model
        try:
            self.figures: dict = model.figures
        except Exception:
            self.figures = None

    def predict(self, context, model_input: dict):
        """The wrapped model class must have a `predict` method which returns a dictionary"""
        return self.model.predict(model_input)


class Umlaut:
    """A class for abstracting training and querying models in MLflow"""

    def __init__(
        self,
        model_name: str,
        run_name: str = "Update",
        folder_name: str = None,
        tracking_server: str = None,
    ):
        self.ENVIRONMENT = os.environ.get("ENVIRONMENT")
        self.DB_USERNAME = os.environ.get("DB_USERNAME")
        self.DB_PASSWORD = os.environ.get("DB_PASSWORD")
        self.DB_HOSTNAME = os.environ.get("DB_HOSTNAME")
        self.DB_NAME = os.environ.get("DB_NAME")
        self.DB_PORT = os.environ.get("DB_PORT")
        self.DB_ARTIFACT_TABLE = os.environ.get("DB_ARTIFACT_TABLE")

        self.model_name = model_name
        self.run_name = run_name
        if folder_name:
            self.folder_name = folder_name
        else:
            self.folder_name = str(model_name).replace(" ", "_").strip()
        if tracking_server:
            self.tracking_server = tracking_server
        else:
            self.tracking_server = os.environ.get("UMLAUT_TRACKING_SERVER") or None

        self.model = None

        if self.tracking_server:
            mlflow.set_tracking_uri(
                f"{self.tracking_server}"
            )
            self.artifact_location = f"mlflow-artifacts:/{self.folder_name}"
        else:
            mlflow.set_tracking_uri(
                f"postgresql+psycopg2://{self.DB_USERNAME}:{self.DB_PASSWORD}@{self.DB_HOSTNAME}:{self.DB_PORT}/{self.DB_NAME}?"
                f"options=-csearch_path%3D{self.DB_ARTIFACT_TABLE}"
            )
            self.artifact_location = (
                f"s3://{self.ENVIRONMENT}-ml-artifacts/{self.folder_name}/"
            )

    def track_model(self, model, code_path: list = None):
        """
        Trains a new version of the initiated model and pushes it to MLflow in a new run.
        Once pushed, the model can be associated to an existing model in the MLflow UI.
        :param object model: model to be created or updated
        :param list code_path: A list of local filesystem paths to Python file dependencies (or directories containing
                        file dependencies). These files are prepended to the system path before the model is loaded.
        """
        from mlflow.tracking import MlflowClient

        self.model = model
        mlf_client = MlflowClient()
        experiment = mlf_client.get_experiment_by_name(f"{self.model_name}")
        experiment_id = experiment.experiment_id or mlf_client.create_experiment(
            f"{self.model_name}", artifact_location=self.artifact_location
        )

        with mlflow.start_run(experiment_id=experiment_id, run_name=self.run_name):
            self.model = PyfuncWrapper(self.model)
            mlflow.pyfunc.log_model(
                artifact_path="model",
                python_model=self.model,
                code_path=code_path,
                registered_model_name=f"{self.model_name}",
            )

            with suppress(Exception):
                if self.model.figures:
                    """The model `figures: dict` variable is used for logging Plotly performance plots.
                    All figures must be saved as html files.
                    Format: {"<plot_name>.html": plotly.express plot}
                    """
                    for figure_name in self.model.figures:
                        figure = self.model.figures.get(figure_name)
                        mlflow.log_figure(figure, figure_name)

    def query_model(
        self,
        input_config: dict,
        result_keys: list = None,
        stage: str = "Production",
        nested_run: bool = False,
    ) -> Any:
        """Queries the registered model.
        :param dict input_config: input parameters specific to the model
        :param list result_keys: list of items to be stored in results.txt
        :param str stage: stage of the model to be queried
        :param bool nested_run: whether to include a nested model
        :return Any: the result from the model with varying type {dict, list, tuple, or pd.Dataframe}
        """
        import datetime as dt
        from mlflow.tracking import MlflowClient

        mlf_client = MlflowClient()
        experiment_id = mlf_client.get_experiment_by_name(
            f"{self.model_name}"
        ).experiment_id
        self.model = mlflow.pyfunc.load_model(f"models:/{self.model_name}/{stage}")

        with mlflow.start_run(
            experiment_id=experiment_id, run_name="Query", nested=nested_run
        ):
            result = self.model.predict(input_config)

            mlflow.log_params(
                {
                    "timestamp": dt.datetime.now(),
                    "input_dict": input_config,
                    "model_id": str(self.model.metadata.model_uuid),
                    "model_run_id": str(self.model.metadata.run_id),
                    "model_created": str(self.model.metadata.utc_time_created),
                }
            )

            with suppress(TypeError):
                if result_keys:
                    """Drop any keys not in result_keys"""
                    result = {k: result[k] for k in result_keys if k in result}

            try:
                log_result: dict = {}
                if isinstance(result, list):
                    log_result = {"result": result}
                elif isinstance(result, dict):
                    log_result = result
                elif isinstance(result, tuple):
                    log_result = {y: x for x, y in result}
                elif isinstance(result, pd.DataFrame):
                    log_result = result.to_json(orient="records")
                mlflow.log_text(str(log_result), "results.json")
                with suppress(AttributeError, mlflow.exceptions.MlflowException):
                    """Only log numeric metrics"""
                    mlflow.log_metrics(log_result)
            except AttributeError as e:
                mlflow.log_text(str({"Error": e}), "results.json")

        return result

    def track_dataset(
        self,
        dataset: pd.DataFrame,
        edp_location: str = None,
        initial_query: str = None,
        transformations: list = None,
        date_field: str = None,
        dataset_name: str = None,
        run_name: str = "Update",
        custom_params: dict = None,
        custom_metrics: dict = None,
    ):
        """Saves a dataset to MLflow tracking.
        :param pd.DataFrame dataset: dataset to be tracked
        :param str edp_location: the schema.table location in EDP
        :param str initial_query: the initial query used to build the dataset
        :param list transformations: a list of transformations performed on the dataset
        :param str date_field: if the dataset is in time series, used for retrieving only the latest records
        :param str dataset_name: set a custom dataset name, defaults to the same name as the experiment
        :param str run_name: set a custom run name, defaults to "Update"
        :param dict custom_params: log a dictionary of custom params {"param_name": value}
        :param dict custom_metrics: log a dictionary of custom metrics {"metric_name": value}
        """
        from mlflow.tracking import MlflowClient
        import datetime as dt

        mlf_client = MlflowClient()
        experiment = mlf_client.get_experiment_by_name(f"{self.model_name}")
        experiment_id: str = (
            experiment.experiment_id
            if experiment
            else mlf_client.create_experiment(
                self.model_name, artifact_location=self.artifact_location
            )
        )

        with mlflow.start_run(experiment_id=experiment_id, run_name=run_name):
            mlflow.log_params(
                {
                    "timestamp": dt.datetime.now(),
                    "dataset": dataset_name or self.model_name,
                    "location": edp_location or None,
                    "length": len(dataset),
                }
            )
            with suppress(Exception):
                mlflow.log_param(
                    {"latest_records": dataset[date_field].max() if date_field else None}
                )
            with suppress(Exception):
                if custom_params:
                    mlflow.log_params(custom_params)
            with suppress(Exception):
                if custom_metrics:
                    mlflow.log_metrics(custom_metrics)
            with suppress(Exception):
                mlflow.log_text(
                    dataset.to_csv(),
                    artifact_file=f"{self.model_name}_{dt.date.today()}.csv",
                )
            if date_field:
                with suppress(Exception):
                    latest = dataset[
                        dataset[date_field] >= dataset[date_field].max()
                    ].reset_index(drop=True)
                    mlflow.log_text(
                        latest.to_csv(),
                        artifact_file=f"latest_{self.model_name}_{dt.date.today()}.csv",
                    )
            if initial_query:
                with suppress(Exception):
                    mlflow.log_text(
                        initial_query, artifact_file=f"{self.model_name}_query.txt"
                    )
            if transformations:
                with suppress(Exception):
                    mlflow.log_text(
                        json.dumps(transformations),
                        artifact_file=f"{self.model_name}_transformations.json",
                    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['umlaut', 'umlaut.model_helpers']

package_data = \
{'': ['*']}

install_requires = \
['mlflow==1.28.0']

setup_kwargs = {
    'name': 'umlaut',
    'version': '0.0.1',
    'description': 'Umlaut is a library for training and querying ML models',
    'long_description': '# uMLaut\n\nThe uMLaut library simplifies model deployment and querying. It provides a single\naccess point for all of your organizations models and an interface to interact with all of them in the same way. Umlaut `models` can be as extensive as deep learning models or as simple as a reusable code block.\n\n- Simple model lifecycle management\n- Easily maintain and access multiple versions of the same model\n- Quickly share business logic in reusable modules\n- User interface with `MLflow`\n- Audit tracking history (roadmap)\n- Auto-deployed models that can be queried through an API (roadmap)\n\n____\n## Umlaut Class\nA Python class to assist with saving and querying business logic.\n\n- `track_model`: Converts a block of business logic into an Umlaut compatible `model`\n- `query_model`: Queries a previously trained `model` and saves audit metadata\n- `track_dataset`: Saves reporting datasets along with the initial query and underlying data that built it (roadmap)\n- `audit_model (roadmap)`: Retrieve the results of a model run for a historic date\n- `audit_dataset (roadmap)`: Retrieve a dataset as it was on a historic date\n\n### Developing models with Umlaut\nCustom `models` can be saved from any repository. Ensure the code block is in a Python `Class` and follow the example below.\n\n```\nfrom umlaut.core import Umlaut\n\nclass ExampleModel():\n    """Example business logic that can be wrapped into a model.\n       The class _must_ contain a \'predict\' method.\n    """\n    def business_logic(self, record: dict) -> bool:\n        if record.get("sales") > 5:\n            return True\n        else:\n            return False\n\n    def predict(self, model_input: dict) -> bool:\n        return self.business_logic()\n\nif __name__ == "__main__":\n    from umlaut.core import Umlaut\n\n    model = Umlaut(model_name="example model")\n    model.track_model(ExampleModel())\n```\n\nThis will push the latest changes of `ExampleModel()` to MLflow as a new model version. Navigate to the MLflow Tracking Server to find the latest push and associate it to the MLflow model.\n',
    'author': 'Andrew Dunkel',
    'author_email': 'andrew.dunkel1@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/andrewdunkel/uMLaut',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.12,<4.0',
}


setup(**setup_kwargs)
